/*
  Description  : Provide common functions for drawing, distance calculating, and etc. 
                 Store the constant values. 
  Developer    : Simen
  Contact      : SimenWu@hotmail.com
  Last Modified: 03/02/2003
*/

#include "StdAfx.h"
#include "util.h"

UnitInfo Units[UNIT_NUMS]= {
	{0, 1.0, "Unknown", "Unknown"},
	{9001, 1.0, "Meters", "m"},
	{9003, 0.30480061, "Feet", "ft"},
	{9012, 0.914398415, "Yards", "yd"},
	{9013, 0.914398531, "Yards", "yd"},
	{9035, 1609.347219, "Miles", "mi"},
	{9036, 1000.0, "Kilometers", "km"},
	{9102, 0.0175, "Degrees", "d"}
};

int PixelsPerInchForScreen()
{
  HDC hdc=NULL;
  int n=0;
  hdc=GetDC(NULL);
  n=GetDeviceCaps(hdc, LOGPIXELSX);  
  if (hdc)
	ReleaseDC(NULL, hdc);
  return n;
}

//for multipliers ranging from .00001 to 10000000000
//  Candidates are 1, 2, 2.5, and 5 * multiplier
//    First candidate >starting interval is new interval 
//  next candidate
//next multiplier
double GetRoundIncrement(double StartValue)
{
  int nPower; //power of 10. Range of -5 to 10 gives huge scale range.
  double dMultiplier; //Mulitiplier, =10^exp, to apply to nice numbers.
  double dCandidate;  //Candidate value for new interval.
  for (nPower=PowerRangeMin; nPower<=PowerRangeMax; nPower++) {
    dMultiplier=pow(10, nPower);
	for (int i=0; i<nNiceNumber; i++) {
  	  dCandidate=NiceNumberArray[i] * dMultiplier;
	  if (dCandidate > StartValue)
        return dCandidate;
	}
  }
  return dCandidate; //return the maximum
}

//Calculate the Representative Fraction Scale for non Lat/Long map.
//MapUnitFactor is the factor the unit used on the map
double CalcRFScale(double widthMap, double widthPage, double MapUnitFactor)
{
  double fMapWidth, fPageWidth;
  int nPxlPerInch=PixelsPerInchForScreen();
  double ratio;

  if (widthPage<=0)
    return 0.0;
//convert map width to meters
  fMapWidth=widthMap * MapUnitFactor;
//convert page width to meters.
  try {
    fPageWidth=widthPage / ((double)nPxlPerInch) * metersPerInch;
    ratio=fabs(fMapWidth / fPageWidth);
  }
  catch (...) {
    ratio=0.0;
  }
  return ratio;
}

//Calculate the Representative Fraction Scale for Lat/Long map.
double CalcRFScaleD(double lon1, double lon2, double lat, double widthPage)
{
  double distance=GC_Range(lon1, lon2, lat);
  double r;
  r=CalcRFScale(distance, widthPage, 1);
  return r;
}

void DrawRectangle(HDC hdc, RECT r, COLORREF cr, int nWidth, int penStyle)
{
  HPEN hOldPen, hPen;  
//store the old settings.
  hPen=CreatePen(penStyle, nWidth, cr);
  hOldPen=(HPEN)SelectObject(hdc, hPen);

//Draw
  POINT pt;
  POINT vertices[5];
  pt.x=r.left;  pt.y=r.top;    vertices[0]=pt;
  pt.x=r.left;  pt.y=r.bottom; vertices[1]=pt;
  pt.x=r.right; pt.y=r.bottom; vertices[2]=pt;
  pt.x=r.right; pt.y=r.top;    vertices[3]=pt;
  pt.x=r.left;  pt.y=r.top;    vertices[4]=pt;

  Polyline(hdc, vertices, 5);

//Restore the old settings.
  SelectObject(hdc, hOldPen);
  DeleteObject(hPen);
}

void FillRectangle(HDC hdc, RECT r, int brStyle, COLORREF crBrush, int penStyle, COLORREF crPen, int nPenWidth)
{
  HPEN hOldPen, hPen;
  HBRUSH hOldBrush, hBrush;
  LOGBRUSH lb;

  lb.lbColor=crBrush;
  lb.lbStyle=brStyle;
  
  hPen=CreatePen(penStyle, nPenWidth, crPen);
  hOldPen=(HPEN)SelectObject(hdc, hPen);
  hBrush = CreateBrushIndirect(&lb);   
  hOldBrush = (HBRUSH)SelectObject(hdc, hBrush);

  Rectangle(hdc, r.left, r.top, r.right, r.bottom);

  SelectObject(hdc, hOldPen);
  SelectObject(hdc, hOldBrush);
  DeleteObject(hPen);
  DeleteObject(hBrush);
}

double DiffLongitude(double lon1, double lon2)
{
  double diff;

  if (lon1>180.0)
    lon1 = 360.0 - lon1;
  if (lon2>180.0)
    lon2 = 360.0 - lon2;

  if ((lon1>=0.0) && (lon2>=0.0))
    diff = lon2-lon1;
  else
    if ((lon1< 0.0) && (lon2< 0.0))
      diff = lon2-lon1;
    else { // different hemispheres
      if (lon1<0)
        lon1 = -1*lon1;
      if (lon2<0)
        lon2 = -1*lon2;
      diff = lon1 + lon2;
      if (diff>180.0)
        diff = 360.0 - diff;
    }
  return diff;
}

//---------------------------------------------------------------------------
// Calcualte the distance between 2 points: (lon1,lat1) and (lon2,lat2).
//---------------------------------------------------------------------------
double GC_Range(double lon1, double lat1, double lon2, double lat2)
{
  double arg1, arg2, dLon;
  dLon = DiffLongitude(lon1,lon2); // Delta Longitude in DEG
  arg1 = sin(lat1*DTR)*sin(lat2*DTR);
  arg2 = cos(lat1*DTR)*cos(lat2*DTR)*cos(dLon*DTR);

  return metersPerDegreeAtEquator * acos( arg1+arg2 )/ DTR;
}

//---------------------------------------------------------------------------
// Calcualte the distance between 2 points: (lon1,lat) and (lon2,lat).
//---------------------------------------------------------------------------
double GC_Range(double lon1, double lon2, double lat)
{
  double dLon;
  double distance;
  dLon = DiffLongitude(lon1,lon2); // Delta Longitude in DEG   
  lat=fabs(lat);
  if (lat>=90.0)
	lat=89.999;
  distance=cos(lat*DTR) * metersPerDegreeAtEquator * dLon;
  return distance;
}

void GetUnitInformation(int unit_id, double &factor, char *name, char *short_name)
{
  int index=0;
  for (int i=0; i<UNIT_NUMS; i++) {
    if (unit_id==Units[i].id) { //use the first unit (custom) if it's not found in the pre-defined unit list
	  index=i;
	  break;
    }
  }
//get data
  factor=Units[index].factor;
  strcpy(name, Units[index].name);
  strcpy(short_name, Units[index].short_name);  
}

void CalcBarScale(int nWidthDC, int nNumTics, double fMapScale, double fBarUnitFactor, int &PixelsPerTic, double &SBUnitsPerTic)
{
  double fBarScale;
  double fBarUnitsPerPixel;
  int nMinPixelsPerTic;
  int nPixelsPerInch;
  
  nMinPixelsPerTic=nWidthDC/(nNumTics*2);
  fBarScale=fMapScale / fBarUnitFactor; //scalebar's scale.
  nPixelsPerInch=PixelsPerInchForScreen();
  fBarUnitsPerPixel=fBarScale * metersPerInch / nPixelsPerInch; 

//calculate the result
  SBUnitsPerTic= nMinPixelsPerTic * fBarUnitsPerPixel;
  SBUnitsPerTic=GetRoundIncrement(SBUnitsPerTic);
  PixelsPerTic= (int)(SBUnitsPerTic / fBarUnitsPerPixel);
}

void DrawBarWithStyle(HDC hdc, int nOffsetX, int nOffsetY, int nNumTics, int nTicLength, int nBarWidth, COLORREF crBar1, COLORREF crBar2, BOOL bOutline, COLORREF crOutline, int style)
{
  if (nBarWidth>1) {
    switch (style) {
	  case 0:
        DrawTicBarStandard(hdc, nNumTics, nBarWidth, crBar1, crBar2, nTicLength, nOffsetX, nOffsetY, bOutline, crOutline);
		break;
	  case 1:
        DrawTicBarMeridian(hdc, nNumTics, nBarWidth, crBar1, crBar2, nTicLength, nOffsetX, nOffsetY, bOutline, crOutline);
		break;
	  case 2:
        DrawTicBarMeridian1(hdc, nNumTics, nBarWidth, crBar1, crBar2, nTicLength, nOffsetX, nOffsetY, bOutline, crOutline);
		break;
      default:
        DrawTicBarStandard(hdc, nNumTics, nBarWidth, crBar1, crBar2, nTicLength, nOffsetX, nOffsetY, bOutline, crOutline);
		break;
    }
  }
  else
    DrawTicLine(hdc, nNumTics, crBar1, crBar2, nTicLength, nOffsetX, nOffsetY);
}

void DrawTicLine(HDC hdc, int nTics, COLORREF crBar1, COLORREF crBar2, int nTicLength, int xOrigin, int yOrigin)
{
  HPEN hOldPen, hPen1, hPen2;
  POINT p;
  int x, y;

  hPen1=CreatePen(PS_SOLID, 1, crBar1);
  hPen2=CreatePen(PS_SOLID, 1, crBar2);
  hOldPen=(HPEN)SelectObject(hdc, hPen1);

  x=xOrigin;
  y=yOrigin;
  MoveToEx(hdc, x, y, &p);
  for (int i=0; i<nTics; i++) {
    LineTo(hdc, x, y);
	if (i%2)
      SelectObject(hdc, hPen1);
	else
      SelectObject(hdc, hPen2);
	x +=nTicLength;
  }
  SelectObject(hdc, hOldPen);
  DeleteObject(hPen1);
  DeleteObject(hPen2);
}

//---------------------------------------------------------------------------
// Draw the standard style scalebar tics
//---------------------------------------------------------------------------
void DrawTicBarStandard(HDC hdc, int nTics, int nWidth, COLORREF crBar1, COLORREF crBar2, int nTicLength, int xOrigin, int yOrigin, BOOL bOutline, COLORREF crOutline)
{
  HPEN hOldPen, hPen1, hPen2;
  HBRUSH hOldBrush, hBrush1, hBrush2;
  LOGBRUSH lb1, lb2;
  int x, y1, y2;
  COLORREF cr1, cr2;

//Create pens.
  if (bOutline) {
    cr1=crOutline;
	cr2=crOutline;
  }
  else {
	cr1=crBar1;
	cr2=crBar2;
  }
  hPen1=CreatePen(PS_SOLID, 1, cr1);
  hPen2=CreatePen(PS_SOLID, 1, cr2);
  hOldPen=(HPEN)SelectObject(hdc, hPen1);

//Create brushes.
  lb1.lbColor=crBar1;
  lb1.lbStyle=BS_SOLID;
  lb2.lbColor=crBar2;
  lb2.lbStyle=BS_SOLID;
  hBrush1 = CreateBrushIndirect(&lb1);   
  hBrush2 = CreateBrushIndirect(&lb2);   
  hOldBrush = (HBRUSH)SelectObject(hdc, hBrush1);

  x=xOrigin;
  y1=yOrigin;
  y2=y1+nWidth;
  for (int i=0; i<nTics; i++) {
    Rectangle(hdc, x, y1, x+nTicLength+1, y2);
	if (i%2) {
      SelectObject(hdc, hPen1);
      SelectObject(hdc, hBrush1);
	}
	else {
      SelectObject(hdc, hPen2);
      SelectObject(hdc, hBrush2);
	}
	x +=nTicLength;
  }

//Restore old setttings.
  SelectObject(hdc, hOldPen);
  SelectObject(hdc, hOldBrush);
  DeleteObject(hPen1);
  DeleteObject(hPen2);
  DeleteObject(hBrush1);
  DeleteObject(hBrush2);
}

//---------------------------------------------------------------------------
// Draw the Meridian style scalebar tics
//---------------------------------------------------------------------------
void DrawTicBarMeridian(HDC hdc, int nTics, int nWidth, COLORREF crBar1, COLORREF crBar2, int nTicLength, int xOrigin, int yOrigin, BOOL bOutline, COLORREF crOutline)
{
  HPEN hOldPen, hPen1, hPen2;
  HBRUSH hOldBrush, hBrush1, hBrush2;
  LOGBRUSH lb1, lb2;
  int x, y1, y2, y12;
  COLORREF cr1, cr2;

//Create pens.
  if (bOutline) {
    cr1=crOutline;
	cr2=crOutline;
  }
  else {
	cr1=crBar1;
	cr2=crBar2;
  }
  hPen1=CreatePen(PS_SOLID, 1, cr1);
  hPen2=CreatePen(PS_SOLID, 1, cr2);
  hOldPen=(HPEN)SelectObject(hdc, hPen1);

//Create brushes.
  lb1.lbColor=crBar1;
  lb1.lbStyle=BS_SOLID;
  lb2.lbColor=crBar2;
  lb2.lbStyle=BS_SOLID;
  hBrush1 = CreateBrushIndirect(&lb1);   
  hBrush2 = CreateBrushIndirect(&lb2);   
  hOldBrush = (HBRUSH)SelectObject(hdc, hBrush1);

  x  =xOrigin;
  y1 =yOrigin;
  y2 =y1+nWidth;
  y12=y1+nWidth/2;
  for (int i=0; i<nTics; i++) {
	if (i%2) {   
      SelectObject(hdc, hPen2);
      SelectObject(hdc, hBrush2);
      Rectangle(hdc, x, y1, x+nTicLength+1, y12+1);

      SelectObject(hdc, hPen1);
      SelectObject(hdc, hBrush1);
      Rectangle(hdc, x, y12, x+nTicLength+1, y2);
    }
	else {
      SelectObject(hdc, hPen1);
      SelectObject(hdc, hBrush1);
      Rectangle(hdc, x, y1, x+nTicLength+1, y12+1);

      SelectObject(hdc, hPen2);
      SelectObject(hdc, hBrush2);
      Rectangle(hdc, x, y12, x+nTicLength+1, y2);
	}
	x +=nTicLength;
  }

//Restore old setttings.
  SelectObject(hdc, hOldPen);
  SelectObject(hdc, hOldBrush);
  DeleteObject(hPen1);
  DeleteObject(hPen2);
  DeleteObject(hBrush1);
  DeleteObject(hBrush2);
}

//---------------------------------------------------------------------------
// Draw the Meridian1 style scalebar tics
//---------------------------------------------------------------------------
void DrawTicBarMeridian1(HDC hdc, int nTics, int nWidth, COLORREF crBar1, COLORREF crBar2, int nTicLength, int xOrigin, int yOrigin, BOOL bOutline, COLORREF crOutline)
{
  HPEN hOldPen, hPen1, hPen2;
  HBRUSH hOldBrush, hBrush1, hBrush2;
  LOGBRUSH lb1, lb2;
  int x, y1, y2, y12;
  COLORREF cr1, cr2;

//Create pens.
  if (bOutline) {
    cr1=crOutline;
	cr2=crOutline;
  }
  else {
	cr1=crBar1;
	cr2=crBar2;
  }
  hPen1=CreatePen(PS_SOLID, 1, crOutline);
  hPen2=CreatePen(PS_SOLID, 1, crOutline);
  hOldPen=(HPEN)SelectObject(hdc, hPen1);

//Create brushes.
  lb1.lbColor=crBar1;
  lb1.lbStyle=BS_SOLID;
  lb2.lbColor=crBar2;
  lb2.lbStyle=BS_SOLID;
  hBrush1 = CreateBrushIndirect(&lb1);   
  hBrush2 = CreateBrushIndirect(&lb2);   
  hOldBrush = (HBRUSH)SelectObject(hdc, hBrush1);

  x  =xOrigin;
  y1 =yOrigin;
  y2 =y1+nWidth;
  y12=y1+nWidth/2;
  for (int i=0; i<nTics; i++) {
	if (i%2) {   
      SelectObject(hdc, hPen2);
      SelectObject(hdc, hBrush2);
      Rectangle(hdc, x, y1, x+nTicLength+1, y2);
    }
	else {
      SelectObject(hdc, hPen1);
      SelectObject(hdc, hBrush1);
      Rectangle(hdc, x, y1, x+nTicLength+1, y12+1);
      Rectangle(hdc, x, y12, x+nTicLength+1, y2);
	}
	x +=nTicLength;
  }

//Restore old setttings.
  SelectObject(hdc, hOldPen);
  SelectObject(hdc, hOldBrush);
  DeleteObject(hPen1);
  DeleteObject(hPen2);
  DeleteObject(hBrush1);
  DeleteObject(hBrush2);
}

//Keep only 5 (ScalePrecisionDigits) digits of prcision for the scale and return the scale after formatted.
double FormatRealScale(double scale)
{
  int rfMagnitude;
  double factor;
  double RoundedScale;

  rfMagnitude=(int)(log10(scale));
  factor=pow(10, rfMagnitude- ScalePrecisionDigits);
  if (factor!=0)
    RoundedScale=int((scale/factor)+0.5)*factor;
  else
    RoundedScale=int(scale+0.5);
  return RoundedScale;  
}

//Return the precision for the segment text of the scale bar.
int PresitionOfSegmentText(double value)
{
  int precision=0;
  double f;
  if (value <1 && value >0) {
    precision= ceil(fabs(log10(value)));
    f=value * pow(10, precision);
    if (f- int(f) > 0.01)
      precision ++;
  }
  return precision;
}
